# -*- coding: utf-8 -*-
"""CTML_Assignment5.ipynb

Automatically generated by Colaboratory.

Original file is located at
    https://colab.research.google.com/drive/1IT9zRA2Tm-Zttmuk5EdMdWllnRGMnu5y

# Assignment 5

## Computational Techniques for Machine Learning

### Professor Miguel Ángel Medina-Pérez

Gregorio Reyes A01139438

Jessica Salinas A00815456

Alejandro Trejo A00818219
"""

# !pip install brminer
# !pip install scikit-posthocs
# !pip install Orange3
# !pip install requests

"""# Libraries"""

import Orange
import scikit_posthocs as sp
import pandas as pd
import numpy as np
import brminer
import seaborn as sns
sns.set(rc={"figure.figsize":(10, 5)})
import matplotlib.pyplot as plt
from scipy import stats
from sklearn.metrics import roc_auc_score
from sklearn.preprocessing import OneHotEncoder
import os
import requests
from zipfile import ZipFile

from sklearn.preprocessing import MinMaxScaler
from sklearn.preprocessing import StandardScaler
from sklearn.mixture import GaussianMixture
from sklearn.ensemble import IsolationForest
from sklearn.svm import OneClassSVM

"""# Brminer Modifications"""

import math
import random
from sklearn.base import BaseEstimator
from sklearn.metrics.pairwise import euclidean_distances
from sklearn.metrics.pairwise import manhattan_distances
from sklearn.metrics.pairwise import cosine_distances
from sklearn.metrics.pairwise import paired_euclidean_distances

from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from brminer import BRM

class my_BRM(BRM):
  def __init__(self, dissimilarity_measure = euclidean_distances):
    super().__init__(classifier_count=100, bootstrap_sample_percent=100, use_bootstrap_sample_count=False, bootstrap_sample_count=0, use_past_even_queue=False, max_event_count=3, alpha=0.5, user_threshold=95)
    self.dissimilarity_measure = dissimilarity_measure

  def fit(self, X, y = None):
      # Check that X and y have correct shape
      if y is not None:
          X_train, y_train = check_X_y(X, y)
      else:
            X_train = check_array(X)
            
      self._similarity_sum = 0
      self._is_threshold_Computed = False

      self.n_features_in_ = X_train.shape[1]

      if self.n_features_in_ < 1:
          raise ValueError('Unable to instantiate the train dataset - Empty vector')     
      
      # self._scaler = MinMaxScaler() ################
      X_train = pd.DataFrame(X_train)
      # X_train = pd.DataFrame(self._scaler.fit_transform(X_train[X_train.columns]), index=X_train.index, columns=X_train.columns)


      self._max_dissimilarity = math.sqrt(self.n_features_in_)
      self._sd = np.empty(0)
      sampleSize = int(self.bootstrap_sample_count) if (self.use_bootstrap_sample_count) else int(0.01 * self.bootstrap_sample_percent * len(X_train));
      self._centers = np.empty((0, sampleSize, self.n_features_in_))

      list_instances = X_train.values.tolist()
      for i in range(0, self.classifier_count):            
          centers = random.choices(list_instances, k=sampleSize)
          self._centers = np.insert(self._centers, i, centers, axis=0)
          self._sd = np.insert(self._sd, i, 2*(np.mean(self.dissimilarity_measure(centers, centers))/self._max_dissimilarity)**2)

      return self
  
  def score_samples(self, X):
      X = np.array(X)
      # X_test = self._scaler.transform(X)
      X_test = X

      result = []
      batch_size = 100
      for i in range(min(len(X_test), batch_size), len(X_test) + batch_size, batch_size):
          current_X_test = X_test[[j for j in range(max(0, i-batch_size), min(i, len(X_test)))]]

          current_similarity = np.average([np.exp(-np.power(np.amin(self.dissimilarity_measure(current_X_test, self._centers[i]), axis=1)/self._max_dissimilarity, 2)/(self._sd[i])) for i in range(len(self._centers))], axis=0)
      
          result = result + [j for j in list(map(self._evaluate, current_similarity))]

      return result

"""# Functions"""

# Function importing Dataset 
def importdata(trainFile, testFile):
    n_rows = get_n_rows(trainFile)
    train = pd.read_csv(trainFile, sep= ',', header = None, skiprows=n_rows) 
    test = pd.read_csv(testFile, sep= ',', header = None, skiprows=n_rows) 
    return train, test    

# Function to preprocess the dataset 
def preprocess_dataset(train, test): 
    ohe = OneHotEncoder(sparse=True)
    objInTrain = len(train)

    allData = pd.concat([train, test], ignore_index=True, sort =False, axis=0)
    AllDataWihoutClass = allData.iloc[:, :-1]
    AllDataWihoutClassOnlyNominals = AllDataWihoutClass.select_dtypes(include=['object'])
    AllDataWihoutClassNoNominals = AllDataWihoutClass.select_dtypes(exclude=['object'])

    encAllDataWihoutClassNominals = ohe.fit_transform(AllDataWihoutClassOnlyNominals)
    encAllDataWihoutClassNominalsToPanda = pd.DataFrame(encAllDataWihoutClassNominals.toarray())
    
    if AllDataWihoutClassOnlyNominals.shape[1] > 0:
      codAllDataAgain = pd.concat([encAllDataWihoutClassNominalsToPanda, AllDataWihoutClassNoNominals], ignore_index=True, sort =False, axis=1)
    else:
      codAllDataAgain = AllDataWihoutClass

    # Seperating the target variable 
    X_train = codAllDataAgain[:objInTrain]
    y_train = train.values[:, -1]

    X_test = codAllDataAgain[objInTrain:]
    y_test = test.values[:, -1]
    
    return X_train, X_test, y_train, y_test

# Function to make predictions 
def prediction(X_test, clf_object):  
    y_pred = clf_object.score_samples(X_test) 
    return y_pred

def download_zip(zip_url=None):
  if zip_url == None:
    url = 'https://git.disroot.org/gregorio/ctml/raw/branch/master/assignment5/Unsupervised_Anomaly_Detection.zip'
  else:
    url = zip_url
  zip_filename = url.split(sep='/')[-1]
  r = requests.get(url)
  with open(zip_filename, 'wb') as f:
    f.write(r.content)
  return zip_filename

def unzip_zip(zip_filename='Unsupervised_Anomaly_Detection.zip'):
  with ZipFile(zip_filename, 'r') as zip: 
    zip.extractall() 
  return zip_filename[:-4]

def get_datasets():
  dataset_list = []
  zip_filename = download_zip()
  extracted_folder = unzip_zip()
  cwd = os.getcwd()
  directory = cwd + '/' + extracted_folder
  directory = 'Unsupervised_Anomaly_Detection/'
  folders = get_filenames(directory)
  for folder in folders:
    dataset_list.append(get_filenames(folder))
  return dataset_list


def get_filenames(directory):
    file_list = []
    for filename in os.listdir(directory):
        f = os.path.join(directory, filename)
        file_list.append(f)
    return file_list

def get_n_rows(filename):
  f = open(filename,"r")
  lines = f.readlines()
  count = 0
  flag = True
  while flag:
    if lines[count][0] == '@':
      count += 1
    else:
      flag = False
  return count

def get_classifiers():
  return [['BRM_euclidean', my_BRM(dissimilarity_measure=euclidean_distances)],\
          ['BRM_manhattan', my_BRM(dissimilarity_measure=manhattan_distances)],\
          ['BRM_cosine', my_BRM(dissimilarity_measure=cosine_distances)],\
          ['GMM', GaussianMixture()],\
          ['ISOF', IsolationForest(random_state=1)],\
          ['ocSVM', OneClassSVM()]]

def get_scores_array(classifiers):
  classifiers_scores = []
  for classifier in classifiers:
    classifiers_scores.append([])
  return classifiers_scores

def run_classifiers(n_datasets=3, data_transformation=None):
  datasets = get_datasets()
  dataset_count = 0
  classifiers = get_classifiers()
  classifiers_scores = get_scores_array(classifiers)
  while dataset_count < n_datasets:
    if datasets[dataset_count][0][-7:] == 'tra.dat':
      trainFile, testFile = datasets[dataset_count][0], datasets[dataset_count][1]
    else:
      trainFile, testFile = datasets[dataset_count][1], datasets[dataset_count][0]
    dataset_count += 1
  
    classifiers_scores = get_auc_scores(classifiers, classifiers_scores, trainFile, testFile, data_transformation)

  scores_dictionary = save_scores_in_dict(classifiers, classifiers_scores)
  return scores_dictionary

def apply_data_transformation(X_train, X_test, data_transformation):
  if data_transformation == None:
    return X_train, X_test
  elif data_transformation == "MinMax":
    min_max_scaler = MinMaxScaler()
    X_train = min_max_scaler.fit_transform(X_train)
    X_test = min_max_scaler.transform(X_test)
    return X_train, X_test
  elif data_transformation == "Standard":
    scaler = StandardScaler()
    X_train = scaler.fit_transform(X_train)
    X_test = scaler.transform(X_test)
    return X_train, X_test

def get_auc_scores(classifiers, classifiers_scores, trainFile, testFile, data_transformation):
  train, test = importdata(trainFile, testFile)
  X_train, X_test, y_train, y_test = preprocess_dataset(train, test)
  X_train, X_test = apply_data_transformation(X_train, X_test, data_transformation)
  classifiers = get_classifiers()
  classifiers_count = 0
  for classifier in classifiers:
    clf = classifier[1]
    clf.fit(X_train, y_train)
    y_pred = clf.score_samples(X_test)
    auc = roc_auc_score(y_test,  y_pred)
    if auc > 0.5:
      classifiers_scores[classifiers_count].append(auc)
    else:
      classifiers_scores[classifiers_count].append(1 - auc)
    classifiers_count += 1

  return classifiers_scores

def save_scores_in_dict(classifiers, scores):
  scores_dictionary = {}
  for i, classifier in enumerate(classifiers):
    scores_dictionary[classifier[0]] = scores[i]
  return scores_dictionary

def get_scores_dataframe(scores_dictionary):
  algorithm = []
  AUC_scores = []
  for key, values in scores_dictionary.items():
    for value in values:
      algorithm.append(key)
      AUC_scores.append(value)
  df = pd.DataFrame()
  df['Algorithm'] = algorithm
  df['AUC'] = AUC_scores
  return df

def create_boxplot(scores_dataframe):
  sns.boxplot(x = 'Algorithm', y = 'AUC', data = scores_dataframe)
  # plt.savefig('boxplot.png',bbox_inches='tight')
  plt.title('Boxplot of Algorithms Performances')
  plt.show()

def friedman_test(scores_dictionary):
  print('\nFriedman Test')
  print(stats.friedmanchisquare(*[values for keys, values in scores_dictionary.items()]))

def posthoc_nemenyi(scores_dictionary):
  data = []
  for key, values in scores_dictionary.items():
    data.append(values)
  data = np.array(data)
  print('\nPosthoc Test (Nemenyi Test)')
  print(sp.posthoc_nemenyi_friedman(data.T))

def get_rank_dataframe(scores_dictionary):
  df = pd.DataFrame()
  for i in range(0,len(scores_dictionary['BRM_euclidean'])):  
    algorithm = []
    AUC_scores = []
    for key in scores_dictionary.keys():
      algorithm.append(key)
      AUC_scores.append(scores_dictionary[key][i])
    df2 = pd.DataFrame()
    df2['Algorithm'] = algorithm
    df2['AUC'] = AUC_scores
    df[i] = df2['AUC'].rank(ascending=False)
  df['Algorithm'] = algorithm
  df['avg_rank'] = df.mean(axis=1)
  return df[['Algorithm','avg_rank']]

def cd_diagram(scores_dictionary):
  rank_dataframe = get_rank_dataframe(scores_dictionary)
  names = rank_dataframe['Algorithm']
  avranks =  rank_dataframe['avg_rank']
  cd = Orange.evaluation.compute_CD(avranks, len(scores_dictionary['BRM_euclidean']))
  Orange.evaluation.graph_ranks(avranks, names, cd=cd, width=6, textspace=1.5)
  # plt.savefig('cddiagram.png',bbox_inches='tight')
  plt.title('Critical Difference Diagram of Algorithm Ranks')
  plt.show()

def create_visualizations(scores_dictionary):
  scores_dataframe = get_scores_dataframe(scores_dictionary)
  create_boxplot(scores_dataframe)
  friedman_test(scores_dictionary)
  posthoc_nemenyi(scores_dictionary)
  cd_diagram(scores_dictionary)

def run_experiments(n_datasets=3, data_transformation = None):
  scores_dictionary = run_classifiers(n_datasets, data_transformation)
  create_visualizations(scores_dictionary)

# run_experiments(60, data_transformation = None)

# run_experiments(60, data_transformation = 'MinMax')

# run_experiments(60, data_transformation = 'Standard')

